"""API for private and public keys"""
from base64 import b64encode, b64decode
from datetime import datetime, timedelta
import hashlib
import logging
import os
import random
import string
from typing import Any, Dict, Callable, List, Optional, Tuple
from uuid import (
    UUID,
    uuid4,
)

import attr
from cryptography.hazmat.primitives.asymmetric import (
    padding,
    rsa,
)
from cryptography.hazmat.primitives import (
    hashes,
    serialization,
)
from Crypto.Cipher import AES


logger = logging.getLogger(__name__)


Plaintext = bytes
Fingerprint = str


@attr.s
class ResourcePrivateKey:
    private_key_id: UUID = attr.ib(default=None)  # primary key
    uri: str = attr.ib(default=None)
    passphrase: str = attr.ib(default=None)
    data: bytes = attr.ib(default=None)
    created_at: datetime = attr.ib(default=None)
    expires_at: datetime = attr.ib(default=None)
    fingerprint: str = attr.ib(default=None)
    rescinded: bool = attr.ib(default=None)
    user_supplied: bool = attr.ib(default=None)


@attr.s
class ResourcePublicKey:
    bri: str = attr.ib(default=None)
    public_key: bytes = attr.ib(default=None)
    expires_at: datetime = attr.ib(default=None)
    fingerprint: str = attr.ib(default=None)


def adapt(row_dict: Dict[str, Any]) -> ResourcePrivateKey:
    return ResourcePrivateKey(
        private_key_id=row_dict["private_key_id"],
        uri=row_dict["uri"],
        passphrase=row_dict["passphrase"],
        data=bytes(row_dict["data"]),
        created_at=row_dict["created_at"],
        expires_at=row_dict["expires_at"],
        fingerprint=row_dict["fingerprint"],
        rescinded=row_dict["rescinded"],
        user_supplied=row_dict["user_supplied"],
    )


@attr.s
class PostgreSQLStore:
    """Resource Private Keys"""
    db: Any = attr.ib(default=None)
    symkey: str = attr.ib()

    @symkey.default
    def symkey_(self):
        self.symkey = os.environ["PG_SYMMETRIC_KEY"]

    def set_rescinded(
        self,
        private_key_id: UUID,
        timenow: datetime,
    ) -> None:
        query = """
        UPDATE tbl_martlet_resource_private_keys
        SET rescinded=true, updated_at=%(timenow)s
        WHERE (
            private_key_id=%(private_key_id)s
        );
        """
        vars_ = vars()
        vars_.update(symkey=self.symkey)
        with self.db.get_cursor(commit=True) as cur:
            cur.execute(
                query=query,
                vars=vars_,
            )

    def insert(self, resource_private_key: ResourcePrivateKey) -> None:
        query = """
        INSERT into tbl_martlet_resource_private_keys (
            private_key_id,
            uri,
            passphrase,
            data,
            created_at,
            expires_at,
            fingerprint,
            rescinded,
            user_supplied
        ) values (
            %(private_key_id)s,
            %(uri)s,
            pgp_sym_encrypt(%(passphrase)s, %(symkey)s),
            %(data)s,
            %(created_at)s,
            %(expires_at)s,
            %(fingerprint)s,
            %(rescinded)s,
            %(user_supplied)s
        );
        """
        vars_ = attr.asdict(resource_private_key)
        vars_.update(symkey=self.symkey)
        with self.db.get_cursor(commit=True) as cur:
            cur.execute(
                query=query,
                vars=vars_,
            )

    def query(
        self,
        uri: str,
        fingerprint: str,
    ) -> Optional[ResourcePrivateKey]:
        query = """
        SELECT
            private_key_id,
            uri,
            pgp_sym_decrypt(passphrase::bytea, %(symkey)s::varchar)
              as passphrase,
            data,
            created_at,
            expires_at,
            fingerprint,
            rescinded,
            user_supplied
        FROM tbl_martlet_resource_private_keys WHERE (
            uri=%(uri)s
        ) AND (
            fingerprint=%(fingerprint)s
        );
        """
        vars_ = vars()
        vars_.update(symkey=self.symkey)
        with self.db.get_cursor(as_dict=True) as cur:
            cur.execute(
                query=query,
                vars=vars_,
            )
            row_dict = cur.fetchone()
        return adapt(row_dict) if row_dict else None

    def query_all(
        self,
        uri: str,
    ) -> List[ResourcePrivateKey]:
        query = """
        SELECT
            private_key_id,
            uri,
            pgp_sym_decrypt(passphrase::bytea, %(symkey)s::varchar)
              as passphrase,
            data,
            created_at,
            expires_at,
            fingerprint,
            rescinded,
            user_supplied
        FROM tbl_martlet_resource_private_keys WHERE (
            uri=%(uri)s
        );
        """
        vars_ = vars()
        vars_.update(symkey=self.symkey)
        with self.db.get_cursor(as_dict=True) as cur:
            cur.execute(
                query=query,
                vars=vars_,
            )
            row_dicts = cur.fetchall()
        return [adapt(row_dict) for row_dict in row_dicts]

    def query_latest_active(
        self,
        uri: str,
        clock=datetime.utcnow,
    ) -> Optional[ResourcePrivateKey]:
        query = """
        SELECT
            private_key_id,
            uri,
            pgp_sym_decrypt(passphrase::bytea, %(symkey)s::varchar)
              as passphrase,
            data,
            created_at,
            expires_at,
            fingerprint,
            rescinded,
            user_supplied
        FROM tbl_martlet_resource_private_keys
        WHERE (
            uri=%(uri)s
        ) AND (
            rescinded=false
        ) AND (
            expires_at > %(timenow)s
        )
        ORDER BY expires_at DESC
        LIMIT 1
        ;
        """
        vars_ = dict(
            uri=uri,
            timenow=clock(),
        )
        vars_.update(symkey=self.symkey)
        with self.db.get_cursor(as_dict=True) as cur:
            cur.execute(
                query=query,
                vars=vars_,
            )
            row_dict = cur.fetchone()
        return adapt(row_dict) if row_dict else None

    def select(self, private_key_id: UUID) -> ResourcePrivateKey:
        query = """
        SELECT
            private_key_id,
            uri,
            pgp_sym_decrypt(passphrase::bytea, %(symkey)s::varchar)
              as passphrase,
            data,
            created_at,
            expires_at,
            fingerprint,
            rescinded,
            user_supplied
            FROM tbl_martlet_resource_private_keys WHERE (
                private_key_id=%(private_key_id)s
            );
        """
        vars_ = vars()
        vars_.update(symkey=self.symkey)
        with self.db.get_cursor(as_dict=True) as cur:
            cur.execute(
                query=query,
                vars=vars_,
            )
            row_dicts = cur.fetchall()
        return adapt(row_dicts[0])


def generate_passphrase():
    elements: List[str] = random.choices(
        string.ascii_letters + string.digits,
        k=32,
    )
    passphrase: str = "".join(elements)
    return passphrase


def create_private_key() -> rsa.RSAPrivateKey:
    key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
    )
    return key


def generate_fingerprint(
    public_key: rsa.RSAPublicKey,
) -> str:
    modulus: int = public_key.public_numbers().n
    modulus_s: str = str(modulus)
    modulus_b: bytes = bytes(modulus_s, "utf-8")
    hash = hashlib.sha256(modulus_b)
    return hash.hexdigest()


def serialize_private_key(
    private_key: rsa.RSAPrivateKey,
    passphrase: str,
) -> bytes:
    passphrase_bytes = passphrase.encode("utf-8")
    pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.BestAvailableEncryption(
            passphrase_bytes
        ),
    )
    return pem


def load_private_key(
    data: bytes,
    passphrase: str,
) -> rsa.RSAPrivateKey:
    reloaded_pk = serialization.load_pem_private_key(
        data,
        password=passphrase.encode("utf-8"),
    )
    return reloaded_pk  # type: ignore


def load_public_key(
    data: bytes,
) -> rsa.RSAPublicKey:
    reloaded_pk = serialization.load_pem_public_key(
        data,
    )
    return reloaded_pk  # type: ignore


def fetch_resource_private_key(
    store: PostgreSQLStore,
    bri: str,
    fingerprint: str,
) -> Optional[ResourcePrivateKey]:
    resource_private_key = store.query(
        uri=bri,
        fingerprint=fingerprint,
    )
    return resource_private_key


def fetch_latest_resource_public_key(
    store: PostgreSQLStore,
    bri: str,
) -> Optional[ResourcePublicKey]:
    resource_private_key = store.query_latest_active(
        uri=bri,
    )
    if resource_private_key is None:
        return None
    private_key = load_private_key(
        resource_private_key.data,
        passphrase=resource_private_key.passphrase,
    )
    public_key = private_key.public_key()
    public_key_bytes = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.PKCS1,
    )
    resource_public_key = ResourcePublicKey(
        bri=bri,
        public_key=public_key_bytes,
        expires_at=resource_private_key.expires_at,
        fingerprint=resource_private_key.fingerprint,
    )
    return resource_public_key


def rescind_key(
    store: PostgreSQLStore,
    private_key_id: UUID,
    clock: Callable = datetime.utcnow,
) -> ResourcePrivateKey:
    timenow = clock()
    store.set_rescinded(
        private_key_id=private_key_id,
        timenow=timenow,
    )
    return store.select(private_key_id=private_key_id)


def rotate_key(
    store: PostgreSQLStore,
    bri: str,
    clock: Callable = datetime.utcnow,
    id_generator: Callable = uuid4,
    passphrase_generator: Callable = generate_passphrase,
    life_in_days: int = 90,
) -> ResourcePrivateKey:
    """Generate new key and store"""
    private_key = create_private_key()
    passphrase = passphrase_generator()
    data: bytes = serialize_private_key(private_key, passphrase=passphrase)
    created_at = clock()
    expires_at = created_at + timedelta(days=life_in_days)
    private_key_id = id_generator()
    fingerprint = generate_fingerprint(private_key.public_key())
    resource_private_key = ResourcePrivateKey(
        private_key_id=private_key_id,
        uri=bri,
        passphrase=passphrase,
        data=data,
        created_at=created_at,
        expires_at=expires_at,
        fingerprint=fingerprint,
        rescinded=False,
        user_supplied=False,
    )
    store.insert(
        resource_private_key=resource_private_key,
    )
    return resource_private_key


def decrypt_pki(
    store: PostgreSQLStore,
    bri: str,
    fingerprint: str,
    ciphertext: bytes,
    clock: Callable = datetime.utcnow,
) -> Plaintext:
    """Raises exception if key is invalid (expired/rescinded)"""
    resource_private_key = store.query(uri=bri, fingerprint=fingerprint)
    if resource_private_key is None:
        raise ValueError("no resource private key found")

    if resource_private_key.rescinded is True:
        raise ValueError("key rescinded")

    private_key = load_private_key(
        data=resource_private_key.data,
        passphrase=resource_private_key.passphrase,
    )
    plaintext: bytes = private_key.decrypt(
        ciphertext=ciphertext,
        padding=padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None,
        ),
    )
    return plaintext


def decrypt_symmetric(
    ciphertext: bytes,
    key: bytes,
    nonce: bytes,
    tag: bytes,
) -> Plaintext:
    cipher = AES.new(key, AES.MODE_EAX, nonce=nonce)
    plaintext = cipher.decrypt(ciphertext)
    try:
        cipher.verify(tag)
        print("The message is authentic:", plaintext)
    except ValueError:
        print("Key incorrect or message corrupted")
    return plaintext


def decrypt(
    store: PostgreSQLStore,
    bri: str,
    fingerprint: str,
    rubric_ciphertext: bytes,
    message_ciphertext: bytes,
) -> Plaintext:
    """Decrypt the rubric then decrypt the message symmetrically"""
    rubric_plaintext: bytes = decrypt_pki(
        store=store,
        bri=bri,
        fingerprint=fingerprint,
        ciphertext=rubric_ciphertext,
    )
    cipher_type, key, nonce, tag = unpack_rubric(rubric_plaintext)
    message_plaintext = decrypt_symmetric(
        ciphertext=message_ciphertext,
        key=key,
        nonce=nonce,
        tag=tag,
    )
    return message_plaintext


def fetch_all_resource_private_keys(
    store: PostgreSQLStore,
    bri: str,
) -> List[ResourcePrivateKey]:
    return store.query_all(uri=bri)


def pack_rubric(
    algorithm: bytes,
    key: bytes,
    nonce: bytes,
    tag: bytes,
) -> bytes:
    algorithm_ = b64encode(algorithm).decode("utf-8")
    key_ = b64encode(key).decode("utf-8")
    nonce_ = b64encode(nonce).decode("utf-8")
    tag_ = b64encode(tag).decode("utf-8")
    rubric = f"{algorithm_}:{key_}:{nonce_}:{tag_}"
    rubric_ = bytes(rubric, "utf-8")
    return rubric_


def unpack_rubric(
    rubric: bytes,
) -> Tuple[bytes, bytes, bytes, bytes]:
    rubric_: str = rubric.decode("utf-8")
    algorithm_, key_, nonce_, tag_ = rubric_.split(":")
    algorithm = b64decode(bytes(algorithm_, "utf-8"))
    key = b64decode(bytes(key_, "utf-8"))
    nonce = b64decode(bytes(nonce_, "utf-8"))
    tag = b64decode(bytes(tag_, "utf-8"))
    return algorithm, key, nonce, tag
